import { Address, ContractState, TestContractResult, HexString, ContractFactory, SubscribeOptions, EventSubscription, CallContractParams, CallContractResult, TestContractParams, ContractEvent, ContractInstance } from "@alephium/web3";
export declare namespace GovernanceTypes {
    type Fields = {
        chainId: bigint;
        governanceChainId: bigint;
        governanceEmitterAddress: HexString;
        tokenBridgeFactory: HexString;
        receivedSequence: bigint;
        messageFee: bigint;
        guardianSets: [HexString, HexString];
        guardianSetIndexes: [bigint, bigint];
        previousGuardianSetExpirationTimeMS: bigint;
    };
    type State = ContractState<Fields>;
    type WormholeMessageEvent = ContractEvent<{
        sender: HexString;
        targetChainId: bigint;
        sequence: bigint;
        nonce: HexString;
        payload: HexString;
        consistencyLevel: bigint;
    }>;
    interface CallMethodTable {
        getMessageFee: {
            params: Omit<CallContractParams<{}>, "args">;
            result: CallContractResult<bigint>;
        };
        parseAndVerifyVAA: {
            params: CallContractParams<{
                data: HexString;
                isGovernanceVAA: boolean;
            }>;
            result: CallContractResult<[
                bigint,
                bigint,
                HexString,
                bigint,
                HexString
            ]>;
        };
        parseAndVerifyGovernanceVAAGeneric: {
            params: CallContractParams<{
                vaa: HexString;
                targetSequence: bigint;
                coreModule: bigint;
                action: HexString;
            }>;
            result: CallContractResult<[bigint, bigint, HexString]>;
        };
    }
    type CallMethodParams<T extends keyof CallMethodTable> = CallMethodTable[T]["params"];
    type CallMethodResult<T extends keyof CallMethodTable> = CallMethodTable[T]["result"];
    type MultiCallParams = Partial<{
        [Name in keyof CallMethodTable]: CallMethodTable[Name]["params"];
    }>;
    type MultiCallResults<T extends MultiCallParams> = {
        [MaybeName in keyof T]: MaybeName extends keyof CallMethodTable ? CallMethodTable[MaybeName]["result"] : undefined;
    };
}
declare class Factory extends ContractFactory<GovernanceInstance, GovernanceTypes.Fields> {
    consts: {
        Version: string;
        GuardianSetExpireDuration: bigint;
        CoreModule: bigint;
        ErrorCodes: {
            InvalidEmitChainId: bigint;
            InvalidEmitAddress: bigint;
            InvalidMessageSize: bigint;
            InvalidSequence: bigint;
            InvalidModule: bigint;
            InvalidActionId: bigint;
            InvalidVersion: bigint;
            InvalidGuardianSetIndex: bigint;
            InvalidGuardianSetSize: bigint;
            InvalidSignatureSize: bigint;
            InvalidSignatureGuardianIndex: bigint;
            InvalidSignature: bigint;
            GuardianSetExpired: bigint;
            InvalidTargetChainId: bigint;
            ContractStateMismatch: bigint;
            InvalidRegisterChainMessage: bigint;
            InvalidTokenId: bigint;
            InvalidNonceSize: bigint;
            TokenNotExist: bigint;
            InvalidTransferTargetChain: bigint;
            InvalidDestroyUnexecutedSequenceMessage: bigint;
            InvalidCaller: bigint;
            ArbiterFeeLessThanAmount: bigint;
            InvalidAttestTokenMessage: bigint;
            InvalidPayloadId: bigint;
            InvalidTransferMessage: bigint;
            ExpectRemoteToken: bigint;
            InvalidConsistencyLevel: bigint;
            InvalidUpdateRefundAddressMessage: bigint;
            TransferAmountLessThanMessageFee: bigint;
            InvalidAttestTokenArg: bigint;
            InvalidAttestTokenHandler: bigint;
            NotSupported: bigint;
        };
        ActionId: {
            ContractUpgrade: string;
            NewGuardianSet: string;
            NewMessageFee: string;
            TransferFee: string;
        };
    };
    at(address: string): GovernanceInstance;
    tests: {
        getMessageFee: (params: Omit<TestContractParams<GovernanceTypes.Fields, never>, "testArgs">) => Promise<TestContractResult<bigint>>;
        publishWormholeMessage: (params: TestContractParams<GovernanceTypes.Fields, {
            payer: Address;
            targetChainId: bigint;
            sequence: bigint;
            nonce: HexString;
            payload: HexString;
            consistencyLevel: bigint;
        }>) => Promise<TestContractResult<null>>;
        parseAndVerifyVAA: (params: TestContractParams<GovernanceTypes.Fields, {
            data: HexString;
            isGovernanceVAA: boolean;
        }>) => Promise<TestContractResult<[bigint, bigint, HexString, bigint, HexString]>>;
        parseAndVerifyGovernanceVAAGeneric: (params: TestContractParams<GovernanceTypes.Fields, {
            vaa: HexString;
            targetSequence: bigint;
            coreModule: bigint;
            action: HexString;
        }>) => Promise<TestContractResult<[bigint, bigint, HexString]>>;
        parseAndVerifyGovernanceVAA: (params: TestContractParams<GovernanceTypes.Fields, {
            vaa: HexString;
            action: HexString;
        }>) => Promise<TestContractResult<[bigint, HexString]>>;
        getGuardiansInfo: (params: TestContractParams<GovernanceTypes.Fields, {
            guardianSetIndex: bigint;
        }>) => Promise<TestContractResult<HexString>>;
        submitContractUpgrade: (params: TestContractParams<GovernanceTypes.Fields, {
            vaa: HexString;
        }>) => Promise<TestContractResult<null>>;
        updatePreviousGuardianSet: (params: Omit<TestContractParams<GovernanceTypes.Fields, never>, "testArgs">) => Promise<TestContractResult<null>>;
        submitNewGuardianSet: (params: TestContractParams<GovernanceTypes.Fields, {
            vaa: HexString;
        }>) => Promise<TestContractResult<null>>;
        submitSetMessageFee: (params: TestContractParams<GovernanceTypes.Fields, {
            vaa: HexString;
        }>) => Promise<TestContractResult<null>>;
        submitTransferFees: (params: TestContractParams<GovernanceTypes.Fields, {
            vaa: HexString;
        }>) => Promise<TestContractResult<null>>;
    };
}
export declare const Governance: Factory;
export declare class GovernanceInstance extends ContractInstance {
    constructor(address: Address);
    fetchState(): Promise<GovernanceTypes.State>;
    getContractEventsCurrentCount(): Promise<number>;
    subscribeWormholeMessageEvent(options: SubscribeOptions<GovernanceTypes.WormholeMessageEvent>, fromCount?: number): EventSubscription;
    methods: {
        getMessageFee: (params?: Omit<CallContractParams<{}>, "args"> | undefined) => Promise<GovernanceTypes.CallMethodResult<"getMessageFee">>;
        parseAndVerifyVAA: (params: GovernanceTypes.CallMethodParams<"parseAndVerifyVAA">) => Promise<GovernanceTypes.CallMethodResult<"parseAndVerifyVAA">>;
        parseAndVerifyGovernanceVAAGeneric: (params: GovernanceTypes.CallMethodParams<"parseAndVerifyGovernanceVAAGeneric">) => Promise<GovernanceTypes.CallMethodResult<"parseAndVerifyGovernanceVAAGeneric">>;
    };
    multicall<Calls extends GovernanceTypes.MultiCallParams>(calls: Calls): Promise<GovernanceTypes.MultiCallResults<Calls>>;
}
export {};
